import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.*;

import org.freepascal.rtl.TObject;

import SecureBlackbox.Base.*;
import SecureBlackbox.SSHCommon.*;
import SecureBlackbox.SSHServer.*;

public class SSHSession extends TObject implements ActionListener {
	private TElSocket m_Socket = null;
	private java.lang.Thread m_Thread = null; 
	private javax.swing.Timer m_SocketSessionTimer = null; 
	private Hashtable<String, Integer> m_authInfo = new Hashtable<String, Integer>();
	private TElSSHServer m_SSHServer = new TElSSHServer();
	private TElSSHMemoryKeyStorage m_HostKeys = new TElSSHMemoryKeyStorage();
	private Object m_Data = null;
	private String m_Status = "";
	private String m_Username = "";
	private String m_Host = "";
	private String m_ClientSoftware = "";
	private Date m_StartTime;
	private boolean m_Error = false;
	private Object m_lock = new Object();
    private boolean m_Opened;
    private SessionEventListener m_eListener;
    
    public Object getData() {
    	return m_Data;
    }
    
    public void setData(Object value) {
    	m_Data = value;
    }
    
    public String getStatus() {
    	return m_Status;
    }
    
    public String getUsername() {
    	return m_Username;
    }
    
    public String getHost() {
    	return m_Host;
    }
    
    public String getClientSoftware() {
    	return m_ClientSoftware;
    }
    
    public Date getStartTime() {
    	return m_StartTime;
    }
    
	public SSHSession(TElSocket socket, int id, SessionEventListener el)
	{
		m_eListener = el;
		m_Socket = socket;
		try
		{
			m_Socket.getNativeSocket().setSoLinger(true, 300);
		}
		catch(Exception exc) 
		{
			Logger.Log("SSHSession() : " + exc.getMessage());
		}

		m_Host = socket.getRemoteAddress() + ":" + socket.getPort();
		m_StartTime = new Date();

		m_SocketSessionTimer = new javax.swing.Timer(5 * 60 * 1000, this);
		m_SocketSessionTimer.start();			

		setupServer();

		java.lang.Thread t = new java.lang.Thread(new Runnable() {
			public void run() {
				readThread();
			}
		});
	
		t.setName("SSH Server worker thread");
        m_Opened = true;
		t.start();
	}

	private void readThread()
	{
		try
		{	
			m_Error = false;
			synchronized (m_lock) {
				m_SSHServer.open();
	
				while ((m_Socket != null) && (m_Socket.getState() == SBSocket.issConnected) && (!m_Error)) 
				{
					if (m_Socket.canReceive(1000000)) 
					{
						m_SSHServer.dataAvailable();
					} 
					else
					{
						java.lang.Thread.sleep(50);
					}
				}
			}
		}
		catch(Exception ex)
		{
			if (!(ex instanceof InterruptedException))
			{
				Logger.Log("ReadThread : " + ex.getMessage());
				closeSession();
			}
			else 
			{
				if (m_eListener != null) 
				{
					m_eListener.SessionClosed(this);
					m_Thread = null;
				}	
			}				
		}
	}

	public void closeSession()
	{
        if (m_Opened)
        {
            try
            {
                cleanUp();
                if (m_eListener != null)
                	m_eListener.SessionClosed(this);
            }
            finally
            {
                m_Opened = false;
            }
        }
	}
	
	private void cleanUp()
	{
		try
		{
			if (m_SSHServer.getActive())
				m_SSHServer.close(true);
	
	        doSocketShutdown();
	
			if (m_Thread != null)
			{			
				m_SocketSessionTimer.stop();
				m_Thread.interrupt();										
			}
		} catch(Exception ex){
			ex.printStackTrace();
			Logger.Log(ex.getMessage());
		}
	}

	private void doSocketShutdown()
	{
		try
		{	
			if (m_Socket != null) 
			{
				Logger.Log("Closing the socket");
				m_Socket.close(false);
				m_Socket = null;
			}
		}
		catch(Exception ex)
		{
			Logger.Log("doSocketShutdown : " + ex.getMessage());
		}
	}

	public void start()
	{
		callSessionInfoChanged();
		m_SocketSessionTimer.restart();
	}

	private void setupServer()
	{
		synchronized (Globals.getSettings())
		{
			try
			{
				TElSSHKey Key = new TElSSHKey();
				byte [] BServerKey = SBUtils.bytesOfString(Globals.getSettings().getServerKey());
				int Result = Key.loadPrivateKey(BServerKey, BServerKey.length, "");
				if (Result == 0) 
					m_HostKeys.add(Key);
				else
					Logger.Log("SSHSession.SetupServer - invalid private key " + Result, true);
			} 
			catch(Exception exc) 
			{
				Logger.Log("SSHSession.SetupServer - invalid private key" 
									  + exc.getMessage(), true);
			}
		}

		m_SSHServer.setKeyStorage(m_HostKeys);
		m_SSHServer.getAllowedSubsystems().add("sftp");
		m_SSHServer.setSoftwareName("SSHBlackbox.8");
		m_SSHServer.setForceCompression(Globals.getSettings().getForceCompression());
		
		initEvents();
	}

	// events
	
	private void initEvents() {
		m_SSHServer.setOnAuthAttempt(new TSSHAuthAttemptEvent(authHandler));
		m_SSHServer.setOnAuthFailed(new TSSHAuthenticationFailedEvent(authFailed));
		m_SSHServer.setOnAuthPassword(new TSSHAuthPasswordEvent(authPwd));
		m_SSHServer.setOnAuthPublicKey(new TSSHAuthPublicKeyEvent(authPubKey));
		m_SSHServer.setOnAuthKeyboard(new TSSHAuthKeyboardEvent(authKeyb));
		m_SSHServer.setOnAuthKeyboardResponse(new TSSHAuthKeyboardResponseEvent(authKeybResp));
		m_SSHServer.setOnFurtherAuthNeeded(new TSSHFurtherAuthNeededEvent(furtherAuth));
		m_SSHServer.setOnSend(new TSSHSendEvent(onSend));
		m_SSHServer.setOnReceive(new TSSHReceiveEvent(onReceive));
		m_SSHServer.setOnCloseConnection(new TSSHCloseConnectionEvent(onCloseConn));
		m_SSHServer.setOnError(new TSSHErrorEvent(onError));
		m_SSHServer.setOnOpenSubsystem(new TSSHOpenSubsystemEvent(onOpenSubs));
		m_SSHServer.setOnOpenShell(new TSSHOpenShellEvent(onOpenShell));
		m_SSHServer.setOnOpenCommand(new TSSHOpenCommandEvent(onOpenCommand));	
	}
	
	TSSHOpenCommandEvent.Callback onOpenCommand = new TSSHOpenCommandEvent.Callback() {
		
		public void tsshOpenCommandEventCallback(TObject sender,
				TElSSHTunnelConnection Connection, String Command) {
			Logger.Log("Command requested. Please complete/rewrite this event. This feature is not released under java environment.");
			m_Status += "Command ";
			callSessionInfoChanged();			
		}
	};
	
	TSSHOpenShellEvent.Callback onOpenShell = new TSSHOpenShellEvent.Callback() {
		
		public void tsshOpenShellEventCallback(TObject sender,
				TElSSHTunnelConnection Connection) {
			Logger.Log("Shell requested. Please complete/rewrite this event. This feature is not released under java environment.");
			m_Status += "Shell ";
			callSessionInfoChanged();	
		}
	};
	
	TSSHOpenSubsystemEvent.Callback onOpenSubs = new TSSHOpenSubsystemEvent.Callback() {
		
		public void tsshOpenSubsystemEventCallback(TObject sender,
				TElSSHTunnelConnection conn, String subsystem) {
			Logger.Log("Subsystem " + subsystem + " opened");
			if (subsystem.compareTo("sftp") == 0) 
			{
				new SFTPSession(conn);
				m_Status += "SFTP ";
				callSessionInfoChanged();	
			}
		}
	};
	
	TSSHErrorEvent.Callback onError = new TSSHErrorEvent.Callback() {
		
		public void tsshErrorEventCallback(TObject sender, int errCode) {
			Logger.Log("SSH protocol error #" + errCode);
			m_Error = true;
		}
	};

	TSSHCloseConnectionEvent.Callback onCloseConn = new TSSHCloseConnectionEvent.Callback() {
		
		public void tsshCloseConnectionEventCallback(TObject sender) {
			closeSession();
			m_Error = true;
		}
	};
	
	TSSHReceiveEvent.Callback onReceive = new TSSHReceiveEvent.Callback() {	
		public void tsshReceiveEventCallback(TObject sender, byte[] buffer, int maxSize, TSBInteger written) {
			written.value = 0;
			try 
			{
				if (m_Socket.canReceive(100000))
				{
					m_Socket.receive(buffer, maxSize, written);
					if (written.value == 0) 
					{
						if (m_SSHServer.getActive())
						{
							m_SSHServer.setOnSend(null);
							synchronized (m_lock) {
								m_SSHServer.close(true);
							}
						}
						else closeSession();
					}
					else
					{
						m_SocketSessionTimer.restart();
					}
				} 
				else 
				{
					written.value = 0;
				}
			} 
			catch(Exception ex)
			{
				written.value = 0;
                Logger.Log("Socket receive operation failed: " + ex.getMessage());
                if (m_SSHServer != null){
					if (m_SSHServer.getActive())
					{
						m_SSHServer.setOnSend(null);
						synchronized (m_lock) {
							m_SSHServer.close(true);
						}
					}
					else closeSession();
                }
			}
		}
		
	};
	
	TSSHSendEvent.Callback onSend = new TSSHSendEvent.Callback() {
		
		public void tsshSendEventCallback(TObject sender, byte[] Buffer) {
			try 
			{
				int toSend = Buffer.length;
				TSBInteger sent = new TSBInteger();
				sent.value = 0;
				int ptr = 0;
				while (toSend > 0) 
				{
					m_Socket.send(Buffer, ptr, toSend, sent);
					ptr += sent.value;
					toSend -= sent.value;
				}
			} 
			catch(Exception ex)
			{
                Logger.Log("Socket send operation failed: " + ex.getMessage());
				if (m_SSHServer.getActive())
				{
					m_SSHServer.setOnSend(null);
					synchronized (m_lock) {
						m_SSHServer.close(true);
						
					}
				}
				else closeSession();
			}
		}
	};
	
	TSSHFurtherAuthNeededEvent.Callback furtherAuth = new TSSHFurtherAuthNeededEvent.Callback() {
		
		public void tsshFurtherAuthNeededEventCallback(TObject sender, String Username, TSBBoolean needed) {
            needed.value = true;
			UserInfo user = Globals.getSettings().FindUser(Username);		
			if (user != null) {
				if (m_authInfo.get(Username) != null) 
				{
					int authFlag = (Integer)m_authInfo.get(Username);
					if ((user.getAuthAll()) && (authFlag == user.getAuthTypes()))
					{
						needed.value = false;
					} 
					else if ((!user.getAuthAll()) && ((authFlag & user.getAuthTypes()) != 0))
					{
						needed.value = false;
					}
				} 
			}
		}
	};
	
	TSSHAuthKeyboardResponseEvent.Callback authKeybResp = new TSSHAuthKeyboardResponseEvent.Callback() {	
		public void tsshAuthKeyboardResponseEventCallback(TObject sender,
				TElStringList requests, TElStringList responses, TSBString name,
				TSBString instruction, TElStringList newRequests, TElBits echoes, TSBBoolean accept) {

			accept.value = false;
			if ((responses != null) && (responses.getCount() == 2)) 
			{
				String Username = responses.getString(0);
				String Password = responses.getString(1);
				UserInfo user = Globals.getSettings().FindUser(Username);
				if (user != null)
				{
					accept.value = (user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_KEYBOARD) > 0;
					accept.value = accept.value & user.PasswordValid(Password);
					if (accept.value) 
					{
						int authFlag;
						if (m_authInfo.get(Username) == null) 
						{
							authFlag = 0;
						} 
						else 
						{
							authFlag = (Integer)m_authInfo.get(Username);
						}
						authFlag = authFlag | SBSSHConstants.SSH_AUTH_TYPE_KEYBOARD;
						if ((user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_PASSWORD) > 0) 
						{
							authFlag = authFlag | SBSSHConstants.SSH_AUTH_TYPE_PASSWORD;
						}
						m_authInfo.put(Username, authFlag);
					}
				} 
			}
		}
	};
	
	TSSHAuthKeyboardEvent.Callback authKeyb = new TSSHAuthKeyboardEvent.Callback() {
		
		public void tsshAuthKeyboardEventCallback(TObject sender, String username,
				TElStringList submethods, TSBString name, TSBString instruction,
				TElStringList requests, TElBits echoes) {
			UserInfo user = Globals.getSettings().FindUser(username);
			if ((user != null) && ((user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_KEYBOARD) > 0))
			{
                name.value = "Keyboard-interactive authentication";
                instruction.value = "Please enter the following information";
				requests.add("Username: ");
				requests.add("Password: ");
				echoes = new TElBits();
				echoes.setSize(2);
				echoes.setBits(0, true);
				echoes.setBits(1, false);
			} 
			
		}
	};
	
	TSSHAuthPublicKeyEvent.Callback authPubKey = new TSSHAuthPublicKeyEvent.Callback() {
		
		public void tsshAuthPublicKeyEventCallback(TObject sender, String username, TElSSHKey key, TSBBoolean accept) {
			UserInfo user = Globals.getSettings().FindUser(username);
			if (user != null)
			{
				accept.value = (user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_PUBLICKEY) > 0;
				accept.value = accept.value & user.KeyValid(key);
				if (accept.value) 
				{
					int authFlag;
					if (m_authInfo.get(username) == null) 
					{
						authFlag = 0;
					} 
					else 
					{
						authFlag = (Integer)m_authInfo.get(username);
					}
					authFlag = authFlag | SBSSHConstants.SSH_AUTH_TYPE_PUBLICKEY;
					m_authInfo.put(username, authFlag);
				}
			} 
			else 
			{
				accept.value = false;
			}
		}
	};
	
	TSSHAuthPasswordEvent.Callback authPwd = new TSSHAuthPasswordEvent.Callback() {
		
		public void tsshAuthPasswordEventCallback(TObject sender, String username,
				String password, TSSHAuthPasswordEventParams params) {
			UserInfo user = Globals.getSettings().FindUser(username);
			if (user != null)
			{
				params.Accept = (user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_PASSWORD) > 0;
				params.Accept = params.Accept & user.PasswordValid(password);
				if (params.Accept) 
				{
					int authFlag;
					if (m_authInfo.get(username) == null) 
					{
                        authFlag = 0;
					} 
					else 
					{
                        authFlag = (Integer)m_authInfo.get(username);
					}
					authFlag = authFlag | SBSSHConstants.SSH_AUTH_TYPE_PASSWORD;
					if ((user.getAuthTypes() & SBSSHConstants.SSH_AUTH_TYPE_KEYBOARD) > 0) 
					{
						authFlag = authFlag | SBSSHConstants.SSH_AUTH_TYPE_KEYBOARD;
					}
					m_authInfo.put(username, authFlag);
				}
			} 
			else 
			{
				params.Accept = false;
			}
		}
	}; 
			
	
	TSSHAuthAttemptEvent.Callback authHandler = new TSSHAuthAttemptEvent.Callback() {
		
		public void tsshAuthAttemptEventCallback(TObject sender, String username, int authType, TSBBoolean accept) {
			UserInfo user = Globals.getSettings().FindUser(username);
			if (user != null)
			{
				accept.value = ((user.getAuthTypes() & authType) > 0);
				m_Username = username;
				m_ClientSoftware = m_SSHServer.getClientSoftwareName();
				callSessionInfoChanged();

			} else{
			accept.value = false;
			}
		}
	};
	
	TSSHAuthenticationFailedEvent.Callback authFailed = new TSSHAuthenticationFailedEvent.Callback() {
	
		public void tsshAuthenticationFailedEventCallback(TObject sender, int authenticationType) {
			Logger.Log("Authentication attempt (" + Globals.AuthTypeToStr(authenticationType) + ") failed", true);
		}
	};
	
	public void actionPerformed(ActionEvent e) {
		onSocketSessionTimeout();
	}

	protected void callSessionInfoChanged() {
		m_eListener.SessionInfoChanged(this);		
	}

	private void onSocketSessionTimeout() {
		m_SocketSessionTimer.stop();
		Logger.Log("Socket closed by timeout");
		closeSession();
	}
}
